# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Microbenchmarks for call_with_retries()."""

from __future__ import (
    annotations,
)
import importlib.util
import logging
import time
import unittest
from typing import (
    Any,
)

from bzfs_main.util.retry import (
    Retry,
    RetryableError,
    RetryError,
    RetryPolicy,
    call_with_retries,
)


#############################################################################
def suite() -> unittest.TestSuite:
    test_cases = [
        TestCallWithRetriesBenchmark,
        TestTenacityBenchmark,
    ]
    return unittest.TestSuite(unittest.TestLoader().loadTestsFromTestCase(test_case) for test_case in test_cases)


#############################################################################
class TestCallWithRetriesBenchmark(unittest.TestCase):

    def test_benchmark_r1k_n0_p0_s0(self) -> None:
        self.benchmark(runs=1000, max_retries=0, max_previous_outcomes=0, success_on=0)

    def test_benchmark_r1k_n1_p0_s1(self) -> None:
        self.benchmark(runs=1000, max_retries=1, max_previous_outcomes=0, success_on=1)

    def test_benchmark_r1k_n2_p0_s2(self) -> None:
        self.benchmark(runs=1000, max_retries=2, max_previous_outcomes=0, success_on=2)

    def test_benchmark_r1k_n1_p1_s1(self) -> None:
        self.benchmark(runs=1000, max_retries=1, max_previous_outcomes=1, success_on=1)

    def test_benchmark_r1k_n0_p0_sinf(self) -> None:
        self.benchmark(runs=1000, max_retries=0, max_previous_outcomes=0)

    def test_benchmark_r1k_n1_p0_sinf(self) -> None:
        self.benchmark(runs=1000, max_retries=1, max_previous_outcomes=0)

    def test_benchmark_r1k_n2_p0_sinf(self) -> None:
        self.benchmark(runs=1000, max_retries=2, max_previous_outcomes=0)

    def test_benchmark_r1k_n5_p0_sinf(self) -> None:
        self.benchmark(runs=1000, max_retries=5, max_previous_outcomes=0)

    @unittest.skip("benchmark; enable for performance comparison")
    def test_xbenchmark_r100k_n1_p1_sinf(self) -> None:
        self.benchmark(runs=100_000, max_retries=1, max_previous_outcomes=1)

    def benchmark(
        self,
        runs: int,
        max_retries: int,
        max_previous_outcomes: int = 0,
        success_on: int = 1_000_000_000,
    ) -> None:
        retry_policy = RetryPolicy(
            max_retries=max_retries,
            min_sleep_secs=0,
            initial_max_sleep_secs=0,
            max_sleep_secs=0,
            max_elapsed_secs=1000,
            max_previous_outcomes=max_previous_outcomes,
            reraise=False,
        )
        retry_policy = retry_policy.copy(timing=retry_policy.timing.copy(on_before_attempt=lambda retry: None))
        iters = 0

        def fn(retry: Retry) -> Any:
            nonlocal iters
            iters += 1
            if retry.count >= success_on:
                return None
            raise RetryableError("fail")

        try:  # warmup
            call_with_retries(fn, policy=retry_policy.copy(max_retries=100), log=None)
        except RetryError:
            pass

        log = logging.getLogger("TestCallWithRetriesBenchmark")
        log.setLevel(logging.INFO)
        if not log.handlers:
            log.addHandler(logging.StreamHandler())

        import gc

        gc.collect()

        iters = 0
        start = time.perf_counter()
        for _ in range(runs):
            try:
                call_with_retries(fn, policy=retry_policy, log=None)
            except RetryError:
                pass

        elapsed_secs = time.perf_counter() - start
        iters_per_sec = float("inf") if elapsed_secs <= 0 else iters / elapsed_secs
        runs_per_sec = float("inf") if elapsed_secs <= 0 else runs / elapsed_secs
        log.info(
            f"call_with_retries benchmark: "
            f"runs={runs}, "
            f"max_retries={max_retries}, "
            f"success_on={success_on}, "
            f"max_previous_outcomes={max_previous_outcomes}, "
            f"runs/sec={runs_per_sec:.0f}, "
            f"iters/sec={iters_per_sec:.0f}, "
            f"elapsed={elapsed_secs:.3f}s"
        )


#############################################################################
@unittest.skipIf(importlib.util.find_spec("tenacity") is None, "tenacity not installed")
class TestTenacityBenchmark(unittest.TestCase):
    """Compare `call_with_retries()` with a tenacity-based version.

    Purpose: Provide a baseline for microbenchmarking.
    Assumptions: tenacity is installed; otherwise the tests are skipped.
    Design: Mirror `TestCallWithRetriesBenchmark` scenarios for comparability.
    """

    def test_benchmark_r1k_n0_p0_s0(self) -> None:
        self.benchmark(runs=1000, max_retries=0, max_previous_outcomes=0, success_on=0)

    def test_benchmark_r1k_n1_p0_s1(self) -> None:
        self.benchmark(runs=1000, max_retries=1, max_previous_outcomes=0, success_on=1)

    def test_benchmark_r1k_n2_p0_s2(self) -> None:
        self.benchmark(runs=1000, max_retries=2, max_previous_outcomes=0, success_on=2)

    def test_benchmark_r1k_n1_p1_s1(self) -> None:
        self.benchmark(runs=1000, max_retries=1, max_previous_outcomes=1, success_on=1)

    def test_benchmark_r1k_n0_p0_sinf(self) -> None:
        self.benchmark(runs=1000, max_retries=0, max_previous_outcomes=0)

    def test_benchmark_r1k_n1_p0_sinf(self) -> None:
        self.benchmark(runs=1000, max_retries=1, max_previous_outcomes=0)

    def test_benchmark_r1k_n2_p0_sinf(self) -> None:
        self.benchmark(runs=1000, max_retries=2, max_previous_outcomes=0)

    def test_benchmark_r1k_n5_p0_sinf(self) -> None:
        self.benchmark(runs=1000, max_retries=5, max_previous_outcomes=0)

    @unittest.skip("benchmark; enable for performance comparison")
    def test_xbenchmark_r100k_n1_p1_sinf(self) -> None:
        self.benchmark(runs=100_000, max_retries=1, max_previous_outcomes=1)

    def benchmark(
        self,
        runs: int,
        max_retries: int,
        max_previous_outcomes: int = 0,
        success_on: int = 1_000_000_000,
    ) -> None:
        from tenacity import RetryError as TenacityRetryError
        from tenacity import (
            Retrying,
            retry_if_exception_type,
            stop_after_attempt,
            wait_none,
        )

        previous_outcomes: tuple[object, ...] = ()

        def before_sleep(retry_state: Any) -> None:
            nonlocal previous_outcomes
            previous_outcomes = previous_outcomes[len(previous_outcomes) - max_previous_outcomes + 1 :] + (
                retry_state.outcome,
            )

        retrying = Retrying(
            retry=retry_if_exception_type(RetryableError),
            stop=stop_after_attempt(101),
            wait=wait_none(),
            before_sleep=before_sleep if max_previous_outcomes > 0 else None,
            reraise=False,
        )
        iters = 0

        def fn() -> None:
            nonlocal iters
            iters += 1
            count = retrying.statistics["attempt_number"] - 1
            if count >= success_on:
                return
            raise RetryableError("fail")

        previous_outcomes = ()
        try:  # warmup
            retrying(fn)
        except TenacityRetryError:
            pass

        retrying = Retrying(
            retry=retry_if_exception_type(RetryableError),
            stop=stop_after_attempt(max_retries + 1),
            wait=wait_none(),
            before_sleep=before_sleep if max_previous_outcomes > 0 else None,
            reraise=False,
        )

        log = logging.getLogger("TestTenacityBenchmark")
        log.setLevel(logging.INFO)
        if not log.handlers:
            log.addHandler(logging.StreamHandler())

        import gc

        gc.collect()

        iters = 0
        start = time.perf_counter()
        for _ in range(runs):
            previous_outcomes = ()
            try:
                retrying(fn)
            except TenacityRetryError:
                pass

        elapsed_secs = time.perf_counter() - start
        iters_per_sec = float("inf") if elapsed_secs <= 0 else iters / elapsed_secs
        runs_per_sec = float("inf") if elapsed_secs <= 0 else runs / elapsed_secs
        log.info(
            f"tenacity benchmark: "
            f"runs={runs}, "
            f"max_retries={max_retries}, "
            f"success_on={success_on}, "
            f"max_previous_outcomes={max_previous_outcomes}, "
            f"runs/sec={runs_per_sec:.0f}, "
            f"iters/sec={iters_per_sec:.0f}, "
            f"elapsed={elapsed_secs:.3f}s"
        )



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Aggregates the various integration test classes into a single `unittest.TestSuite`."""

from __future__ import (
    annotations,
)
import os
import unittest

from bzfs_tests.abstract_testcase import (
    AbstractTestCase,
)
from bzfs_tests.itest.ibase import (
    ParametrizedTestCase,
)
from bzfs_tests.itest.test_incremental_send_steps import (
    IncrementalSendStepsTestCase,
)
from bzfs_tests.itest.test_lima_vm_sh import (
    TestLimaVmScript,
)
from bzfs_tests.itest.test_local import (
    LocalTestCase,
)
from bzfs_tests.itest.test_remote import (
    FullRemoteTestCase,
    MinimalRemoteTestCase,
)
from bzfs_tests.itest.test_smoke_adhoc import (
    AdhocTestCase,
    SmokeTestCase,
)
from bzfs_tests.itest.test_ssh import (
    TestSSHLatency,
    TestSSHMasterIntermittentFailure,
)


def suite() -> unittest.TestSuite:
    ttype = AbstractTestCase()
    suite = unittest.TestSuite()
    suite.addTest(unittest.TestLoader().loadTestsFromTestCase(TestLimaVmScript))
    if not (ttype.is_smoke_test or ttype.is_functional_test or ttype.is_adhoc_test):
        suite.addTest(ParametrizedTestCase.parametrize(IncrementalSendStepsTestCase, {"verbose": True}))
        suite.addTest(ParametrizedTestCase.parametrize(TestSSHLatency))
        suite.addTest(ParametrizedTestCase.parametrize(TestSSHMasterIntermittentFailure))

    # for ssh_mode in ["pull-push"]:
    # for ssh_mode in ["local", "pull-push"]:
    # for ssh_mode in []:
    for ssh_mode in ["local"]:
        for min_pipe_transfer_size in [1024**2]:
            for affix in [""]:
                # no_privilege_elevation_modes = []
                no_privilege_elevation_modes = [False]
                if not (os.getuid() == 0 or ttype.is_smoke_test or ttype.is_functional_test or ttype.is_adhoc_test):
                    no_privilege_elevation_modes.append(True)
                for no_privilege_elevation in no_privilege_elevation_modes:
                    encrypted_datasets = [False]
                    if not (ttype.is_smoke_test or ttype.is_functional_test or ttype.is_adhoc_test):
                        encrypted_datasets += [True]
                    for encrypted_dataset in encrypted_datasets:
                        params = {
                            "ssh_mode": ssh_mode,
                            "verbose": True,
                            "min_pipe_transfer_size": min_pipe_transfer_size,
                            "affix": affix,
                            "skip_missing_snapshots": "continue",
                            "no_privilege_elevation": no_privilege_elevation,
                            "encrypted_dataset": encrypted_dataset,
                        }
                        if ttype.is_smoke_test:
                            suite.addTest(ParametrizedTestCase.parametrize(SmokeTestCase, params))
                        elif ttype.is_adhoc_test:
                            suite.addTest(ParametrizedTestCase.parametrize(AdhocTestCase, params))
                        else:
                            suite.addTest(ParametrizedTestCase.parametrize(LocalTestCase, params))
    if ttype.is_smoke_test or ttype.is_adhoc_test:
        return suite

    # ssh_modes = []
    # ssh_modes = ["pull-push"]
    # ssh_modes = ["local", "pull-push", "push", "pull"]
    # ssh_modes = ["local"]
    ssh_modes = ["local", "pull-push"]
    ssh_modes = ["local"] if ttype.is_functional_test else ssh_modes
    for ssh_mode in ssh_modes:
        min_pipe_transfer_sizes = [0, 1024**2]
        min_pipe_transfer_sizes = [0] if ttype.is_functional_test else min_pipe_transfer_sizes
        for min_pipe_transfer_size in min_pipe_transfer_sizes:
            # for affix in [""]:
            # for affix in ["", ".  -"]:
            for affix in [".  -"]:
                no_privilege_elevation_modes = [False]
                for no_privilege_elevation in no_privilege_elevation_modes:
                    for encrypted_dataset in [False]:
                        params = {
                            "ssh_mode": ssh_mode,
                            "verbose": False,
                            "min_pipe_transfer_size": min_pipe_transfer_size,
                            "affix": affix,
                            "skip_missing_snapshots": "continue",
                            "no_privilege_elevation": no_privilege_elevation,
                            "encrypted_dataset": encrypted_dataset,
                        }
                        suite.addTest(ParametrizedTestCase.parametrize(FullRemoteTestCase, params))

    if os.getuid() != 0 and not ttype.is_functional_test:
        for ssh_mode in ["pull-push", "pull", "push"]:
            for min_pipe_transfer_size in [0]:
                for affix in [""]:
                    for no_privilege_elevation in [True]:
                        # for encrypted_dataset in []:
                        for encrypted_dataset in [False]:
                            params = {
                                "ssh_mode": ssh_mode,
                                "verbose": False,
                                "min_pipe_transfer_size": min_pipe_transfer_size,
                                "affix": affix,
                                "skip_missing_snapshots": "continue",
                                "no_privilege_elevation": no_privilege_elevation,
                                "encrypted_dataset": encrypted_dataset,
                            }
                            suite.addTest(ParametrizedTestCase.parametrize(MinimalRemoteTestCase, params))
    return suite



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Integration tests covering incremental send step selection end-to-end."""

from __future__ import (
    annotations,
)
import subprocess
from typing import (
    final,
)

from bzfs_tests.itest import (
    ibase,
)
from bzfs_tests.itest.ibase import (
    IntegrationTestCase,
    are_bookmarks_enabled,
)
from bzfs_tests.test_incremental_send_steps import (
    TestIncrementalSendSteps,
)
from bzfs_tests.zfs_util import (
    create_bookmark,
    create_filesystem,
    dataset_exists,
    destroy,
    take_snapshot,
)


#############################################################################
@final
class IncrementalSendStepsTestCase(IntegrationTestCase):

    def test_snapshot_series_excluding_hourlies(self) -> None:
        testcase = {None: ["d1", "h1", "d2", "d3", "d4"]}
        expected_results = ["d1", "d2", "d3", "d4"]
        dst_foo = ibase.DST_ROOT_DATASET + "/foo"

        src_foo = create_filesystem(ibase.SRC_ROOT_DATASET, "foo")
        for snapshot in testcase[None]:
            take_snapshot(src_foo, snapshot)
        self.run_bzfs(src_foo, dst_foo, "--quiet", "--include-snapshot-regex", "d.*", "--exclude-snapshot-regex", "h.*")
        self.assert_snapshot_names(dst_foo, expected_results)

        self.tearDownAndSetup()
        src_foo = create_filesystem(ibase.SRC_ROOT_DATASET, "foo")
        for snapshot in testcase[None]:
            take_snapshot(src_foo, snapshot)
        src_snapshot = f"{src_foo}@{expected_results[0]}"
        cmd = f"sudo -n zfs send {src_snapshot} | sudo -n zfs receive -F -u {dst_foo}"  # full zfs send
        subprocess.run(cmd, text=True, check=True, shell=True)
        self.assert_snapshot_names(dst_foo, [expected_results[0]])
        self.run_bzfs(src_foo, dst_foo, "--include-snapshot-regex", "d.*", "--exclude-snapshot-regex", "h.*")
        self.assert_snapshot_names(dst_foo, expected_results)

        self.tearDownAndSetup()
        src_foo = create_filesystem(ibase.SRC_ROOT_DATASET, "foo")
        for snapshot in testcase[None]:
            take_snapshot(src_foo, snapshot)
        src_snapshot = f"{src_foo}@{expected_results[0]}"
        cmd = f"sudo -n zfs send {src_snapshot} | sudo -n zfs receive -F -u {dst_foo}"  # full zfs send
        subprocess.run(cmd, text=True, check=True, shell=True)
        self.assert_snapshot_names(dst_foo, [expected_results[0]])
        if are_bookmarks_enabled("src"):
            create_bookmark(src_foo, expected_results[0], expected_results[0])
            destroy(src_snapshot)
            self.run_bzfs(src_foo, dst_foo, "--include-snapshot-regex", "d.*", "--exclude-snapshot-regex", "h.*")
            self.assert_snapshot_names(dst_foo, expected_results)
            src_snapshot2 = f"{src_foo}@{expected_results[-1]}"
            destroy(src_snapshot2)  # no problem because bookmark still exists
            take_snapshot(src_foo, "d99")
            self.run_bzfs(src_foo, dst_foo, "--include-snapshot-regex", "d.*", "--exclude-snapshot-regex", "h.*")
            self.assert_snapshot_names(dst_foo, expected_results + ["d99"])

        self.tearDownAndSetup()
        src_foo = create_filesystem(ibase.SRC_ROOT_DATASET, "foo")
        for snapshot in testcase[None]:
            take_snapshot(src_foo, snapshot)
        src_snapshot = f"{src_foo}@{expected_results[1]}"  # Note: [1]
        cmd = f"sudo -n zfs send {src_snapshot} | sudo -n zfs receive -F -u {dst_foo}"  # full zfs send
        subprocess.run(cmd, text=True, check=True, shell=True)
        self.assert_snapshot_names(dst_foo, [expected_results[1]])
        self.run_bzfs(
            src_foo,
            dst_foo,
            "--skip-missing-snapshots=continue",
            "--include-snapshot-regex",
            "d.*",
            "--exclude-snapshot-regex",
            "h.*",
        )
        self.assert_snapshot_names(dst_foo, expected_results[1:])

        self.tearDownAndSetup()
        src_foo = create_filesystem(ibase.SRC_ROOT_DATASET, "foo")
        for snapshot in testcase[None]:
            take_snapshot(src_foo, snapshot)
        src_snapshot = f"{src_foo}@{expected_results[1]}"  # Note: [1]
        cmd = f"sudo -n zfs send {src_snapshot} | sudo -n zfs receive -F -u {dst_foo}"  # full zfs send
        subprocess.run(cmd, text=True, check=True, shell=True)
        self.assert_snapshot_names(dst_foo, [expected_results[1]])
        self.run_bzfs(
            src_foo,
            dst_foo,
            "--force",
            "--skip-missing-snapshots=continue",
            "--exclude-snapshot-regex",
            ".*",
        )
        self.assert_snapshot_names(dst_foo, [expected_results[1]])

        self.tearDownAndSetup()
        src_foo = create_filesystem(ibase.SRC_ROOT_DATASET, "foo")
        for snapshot in testcase[None]:
            take_snapshot(src_foo, snapshot)
        src_snapshot = f"{src_foo}@{expected_results[1]}"  # Note: [1]
        cmd = f"sudo -n zfs send {src_snapshot} | sudo -n zfs receive -F -u {dst_foo}"  # full zfs send
        subprocess.run(cmd, text=True, check=True, shell=True)
        self.assert_snapshot_names(dst_foo, [expected_results[1]])
        self.run_bzfs(
            src_foo,
            dst_foo,
            "--force",
            "--skip-missing-snapshots=continue",
            "--include-snapshot-regex",
            "!.*",
        )
        self.assert_snapshot_names(dst_foo, [expected_results[1]])

    def test_snapshot_series_excluding_hourlies_with_permutations(self) -> None:
        for testcase in TestIncrementalSendSteps().permute_snapshot_series(5):
            self.tearDownAndSetup()
            src_foo = create_filesystem(ibase.SRC_ROOT_DATASET, "foo")
            dst_foo = ibase.DST_ROOT_DATASET + "/foo"
            for snapshot in testcase[None]:
                take_snapshot(src_foo, snapshot)
            expected_results = testcase["d"]
            if len(expected_results) > 0 and expected_results[0].startswith("h"):
                expected_results = expected_results[1:]
            # logging.info(f"input   : {','.join(testcase[None])}")
            # logging.info(f"expected: {','.join(expected_results)}")
            self.run_bzfs(
                src_foo,
                dst_foo,
                "--skip-missing-snapshots=continue",
                "--include-snapshot-regex",
                "d.*",
                "--exclude-snapshot-regex",
                "h.*",
            )
            if len(expected_results) > 0:
                self.assert_snapshot_names(dst_foo, expected_results)
            else:
                self.assertFalse(dataset_exists(dst_foo))



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Unit tests for `test_host.sh` input validation.

These tests ensure dangerous `bzfs_test_remote_path` values are rejected before any external commands (rsync/ssh) run.
"""

from __future__ import (
    annotations,
)
import os
import shutil
import stat
import subprocess
import tempfile
import unittest
from pathlib import (
    Path,
)


#############################################################################
def suite() -> unittest.TestSuite:
    test_cases = [
        TestRemotePathValidation,
    ]
    return unittest.TestSuite(unittest.TestLoader().loadTestsFromTestCase(test_case) for test_case in test_cases)


#############################################################################
@unittest.skipIf(shutil.which("bash") is None, "Requires bash")
class TestRemotePathValidation(unittest.TestCase):

    def setUp(self) -> None:
        self.repo_root: Path = Path(__file__).resolve().parents[1]
        self.script: str = os.path.join(self.repo_root, "test_host.sh")

    def test_rejects_dangerous_remote_paths_before_running_rsync(self) -> None:
        # Assert that input validation rejects dangerous paths before rsync/ssh runs.
        invalid_paths: list[str] = [
            "/",
            "//",
            "///",
            ".",
            "..",
            "./x",
            "../x",
            "/.",
            "/..",
            "/./x",
            "/../x",
            "x/.",
            "x/..",
            "x/./y",
            "x/../y",
            "a/.hidden/b",
            "/tmp/bzfs",
            "/tmp/.hidden",
            "/var/tmp/bzfs-tests_1.2.3",
        ]

        for remote_path in invalid_paths:
            with self.subTest(remote_path=remote_path):
                result, marker_content = self._run_test_host_with_stubs(remote_path)
                self.assertEqual(1, result.returncode)
                self.assertIn("error: invalid bzfs_test_remote_path:", result.stderr)
                self.assertIsNone(marker_content)

    def test_allows_reasonable_remote_paths(self) -> None:
        # For valid paths, execution reaches the `rsync` invocation. Since `test_host.sh` runs
        # with `set -e`, the rsync stub's exit code becomes the script's exit code.
        valid_paths: list[str] = [
            "foo",
            "tmp/bzfs",
        ]

        for remote_path in valid_paths:
            with self.subTest(remote_path=remote_path):
                result, marker_content = self._run_test_host_with_stubs(remote_path)
                self.assertEqual(99, result.returncode)
                self.assertEqual("rsync_called\n", marker_content)

    def _run_test_host_with_stubs(self, remote_path: str) -> tuple[subprocess.CompletedProcess[str], str | None]:
        """Runs `test_host.sh` with stubbed rsync/ssh and returns process + marker content."""
        # The test prepends a tmp dir to PATH env var so `test_host.sh` only invokes dummy/stub `rsync`/`ssh` executables.
        # The stubs record whether they were invoked by appending a line to a marker file.
        with tempfile.TemporaryDirectory(prefix="bzfs_test_test_host_") as tmpdir:
            marker_file = Path(os.path.join(tmpdir, "stub_called.txt"))
            self._write_executable_stub(
                Path(os.path.join(tmpdir, "rsync")),
                f"#!/usr/bin/env sh\necho rsync_called >> {marker_file!s}\nexit 99\n",
            )
            self._write_executable_stub(
                Path(os.path.join(tmpdir, "ssh")),
                f"#!/usr/bin/env sh\necho ssh_called >> {marker_file!s}\nexit 98\n",
            )

            env = os.environ.copy()
            env["PATH"] = f"{tmpdir}{os.pathsep}{env.get('PATH', '')}"
            env["bzfs_test_remote_userhost"] = "u@h"
            env["bzfs_test_remote_path"] = remote_path
            env["bzfs_test_remote_private_key"] = "_my_test_dummy_"

            marker_file.unlink(missing_ok=True)
            result = subprocess.run(
                ["bash", str(self.script)],
                check=False,
                cwd=self.repo_root,
                env=env,
                text=True,
                capture_output=True,
            )
            marker_content = marker_file.read_text(encoding="utf-8") if marker_file.exists() else None
            return result, marker_content

    def _write_executable_stub(self, path: Path, content: str) -> None:
        """Writes an executable shell stub to `path` with the given `content`."""
        path.write_text(content, encoding="utf-8")
        path.chmod(path.stat().st_mode | stat.S_IXUSR)



# Copyright 2024 Wolfgang Hoschek AT mac DOT com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Integration tests for SSH behavior used by bzfs."""

from __future__ import (
    annotations,
)
import os
import platform
import pwd
import subprocess
import time
from subprocess import (
    DEVNULL,
    PIPE,
)
from typing import (
    cast,
    final,
)
from unittest.mock import (
    patch,
)

import bzfs_main.replication
from bzfs_main import (
    bzfs,
)
from bzfs_main.configuration import (
    LogParams,
)
from bzfs_main.loggers import (
    get_simple_logger,
)
from bzfs_main.util.connection import (
    SHARED,
    ConnectionPool,
    create_simple_minijob,
    create_simple_miniremote,
)
from bzfs_main.util.utils import (
    get_home_directory,
    getenv_any,
    human_readable_duration,
)
from bzfs_tests.itest import (
    ibase,
)
from bzfs_tests.itest.ibase import (
    SSH_CONFIG_FILE,
    SSH_PROGRAM,
    IntegrationTestCase,
)
from bzfs_tests.tools import (
    gc_disabled,
    stop_on_failure_subtest,
)


#############################################################################
@final
class TestSSHLatency(IntegrationTestCase):

    def run_latency_cmd(self, cmd: list[str], *, close_fds: bool = True) -> tuple[str, str]:
        process = subprocess.run(cmd, stdin=DEVNULL, stdout=PIPE, stderr=PIPE, text=True, check=True, close_fds=close_fds)
        return process.stdout[0:-1], process.stderr[0:-1]  # omit trailing newline char

    def test_ssh_loopback_latency(self) -> None:
        self.setup_basic()
        args = bzfs.argument_parser().parse_args(args=["src", "dst"])
        log_params = LogParams(args)
        log = bzfs_main.loggers.get_logger(
            log_params=log_params, args=args, log=get_simple_logger("test_ssh_loopback_latency")
        )
        p = self.make_params(args=args, log_params=log_params, log=log)

        ssh_opts_list = list(p.src.ssh_extra_opts) + ["-oStrictHostKeyChecking=no"]
        ssh_opts_list += ["-S", os.path.join(p.src.ssh_exit_on_shutdown_socket_dir, "bzfs_test_ssh_socket")]
        ssh_opts_list += ["-p", cast(str, getenv_any("test_ssh_port", "22"))]

        private_key_file2 = os.path.join(get_home_directory(), ".ssh", "testid_rsa")
        src_private_key2 = ["-i", private_key_file2]
        private_key_file = os.path.join(get_home_directory(), ".ssh", "id_rsa")
        src_private_key = ["-i", private_key_file]
        if platform.system() == "Linux":
            ssh_opts_list += src_private_key
        else:
            ssh_opts_list += src_private_key2
        ssh_opts = " ".join(ssh_opts_list)
        memory_hog: str = "x" * 1_000
        # memory_hog: str = "x" * 1_000_000_000

        def run_benchmark(cmd: list[str], check_cmd: list[str]) -> None:
            _dummy = len(memory_hog)
            for close_fds in [False, True]:
                with gc_disabled(run_gc_first=True):
                    iters = 0
                    start_time_nanos = time.monotonic_ns()
                    while time.monotonic_ns() < start_time_nanos + 5 * 1000_000_000:
                        if check_cmd:
                            _stdout, stderr = self.run_latency_cmd(check_cmd, close_fds=close_fds)
                            self.assertIn("Master running", stderr)
                        _stdout, stderr = self.run_latency_cmd(cmd, close_fds=close_fds)
                        iters += 1
                    elapsed = time.monotonic_ns() - start_time_nanos
                t = human_readable_duration(elapsed / iters)
                log.info(f"time/iter: {t}, check: {bool(check_cmd)}, close_fds: {close_fds}, cmd: {' '.join(cmd)}")

        echo_cmd = "echo hello"
        list_cmd = f"{p.zfs_program} list -t snapshot -s createtxg -d 1 -Hp -o guid,name {ibase.SRC_ROOT_DATASET}"
        for base_cmd in [echo_cmd, list_cmd]:
            run_benchmark(p.split_args(base_cmd), check_cmd=[])

        for mode in range(2):
            with stop_on_failure_subtest(i=mode):
                control_persist = 2 if mode == 0 else 2  # noqa: RUF034  # seconds
                master_cmd = p.split_args(f"{SSH_PROGRAM} {ssh_opts} -M -oControlPersist={control_persist}s 127.0.0.1 exit")
                check_cmd = p.split_args(f"{SSH_PROGRAM} {ssh_opts} -O check 127.0.0.1")
                master_is_running = False
                try:
                    log.info(f"mode: {mode}")
                    master_result = self.run_latency_cmd(master_cmd)
                    log.info(f"master result: {master_result}")
                    master_is_running = True
                    if mode == 0:  # test assertions
                        start_time = time.time()
                        while time.time() - start_time < 5:
                            time.sleep(0.5)
                            stdout, stderr = self.run_latency_cmd(check_cmd)
                            log.info(f"check result: {(stdout, stderr)}")
                            self.assertIn("Master running", stderr)
                        log.info("now waiting for expiration of master ...")
                        time.sleep(control_persist + 1)
                        with self.assertRaises(subprocess.CalledProcessError) as cm:
                            self.run_latency_cmd(check_cmd)
                        e = cm.exception
                        log.info(f"check result after expiration: {(e.stdout, e.stderr)}")
                        self.assertNotIn("Master running", e.stderr)
                        self.assertIn("Control socket connect", e.stderr)
                        self.assertIn("No such file or directory", e.stderr)
                        master_is_running = False
                    elif mode >= 1:  # benchmark latency
                        for base_cmd in [echo_cmd, list_cmd]:
                            for check in [False, True]:
                                cmd = p.split_args(f"{SSH_PROGRAM} {ssh_opts} 127.0.0.1 {base_cmd}")
                                run_benchmark(cmd, check_cmd=check_cmd if check else [])
                except subprocess.CalledProcessError as e:
                    log.error(f"error: {(e.stdout, e.stderr)}")
                    raise
                finally:
                    if master_is_running:
                        master_exit_cmd = p.split_args(f"{SSH_PROGRAM} {ssh_opts} -O exit 127.0.0.1")
                        result = self.run_latency_cmd(master_exit_cmd)
                        log.info(f"exit result: {result}")


#############################################################################
@final
class TestSSHMasterIntermittentFailure(IntegrationTestCase):
    """Documents and verifies SSH master behavior under intermittent failure.

    This test exists to answer a subtle but important question for operators: what happens when ``bzfs`` is configured to
    reuse an SSH ControlMaster (``reuse_ssh_connection=True`` with ``ControlPersist``), that master dies unexpectedly, and
    replication continues to run every few seconds? In particular: do SSH commands fail for the remainder of the persistence
    window, or do they keep working, and when is a new master created?

    The expected behavior is:
    * While the master is alive, ``Connection.run_ssh_command`` calls ``refresh_ssh_connection_if_necessary()``, which
      periodically extends the master lifetime by running ``ssh -O check`` and, if needed, starting a new master via
      ``ssh -M -oControlPersist=... exit``. The timestamp ``_last_refresh_time`` records successful refreshes.
    * If the master dies after a refresh, the next ``run_ssh_command`` call detects this and recreates a new ControlMaster.
      From that point on, commands are again multiplexed over a hot TCP connection with low startup latency.
    * If said detection logic would somehow fail (no known case for this exists), then ...
        * subsequent ``run_ssh_command`` calls initially take the fast path:
          they skip the ``-O check`` and directly invoke ``ssh -S <socket> ...``.  OpenSSH then automatically falls back to
          a fresh direct TCP connection when the control socket is missing, so commands continue to succeed, albeit with
          higher latency per call.
        * Once the refresh window has elapsed (ssh_control_persist_secs=600 secs by default), the next
          ``refresh_ssh_connection_if_necessary()`` invocation leaves the fast path, detects that ``-O check`` fails, and
          recreates a new ControlMaster.  From that point on, commands are again multiplexed over a hot TCP connection with
          low startup latency.
    This test method verifies that the actual behavior is as described above even if said detection logic would fail.
    """

    def test_master_dies_then_recovers_via_refresh(self) -> None:
        username = pwd.getpwuid(os.getuid()).pw_name
        ssh_port = int(cast(str, getenv_any("test_ssh_port", "22")))
        args = bzfs.argument_parser().parse_args(args=["src", "dst"])
        log_params = LogParams(args)
        log = bzfs_main.loggers.get_logger(
            log_params=log_params, args=args, log=get_simple_logger("test_ssh_master_intermittent_failure")
        )
        remote = create_simple_miniremote(
            log=log,
            ssh_user_host=f"{username}@127.0.0.1",
            ssh_port=ssh_port,
            ssh_config_file=SSH_CONFIG_FILE,
            ssh_program=SSH_PROGRAM,
            reuse_ssh_connection=True,
            ssh_control_persist_secs=4,  # to speed up the test, reduce 600 sec refresh window down to a few secs
            ssh_control_persist_margin_secs=1,
        )
        pool = ConnectionPool(remote, SHARED)
        try:
            conn = pool.get_connection()
            job = create_simple_minijob(timeout_duration_secs=10.0)
            # Initial command: should create a master and succeed.
            proc1 = conn.run_ssh_command(["echo", "one"], job=job, stdout=PIPE, stderr=PIPE, text=True, check=True)
            self.assertEqual("one\n", proc1.stdout)

            ssh_cmd = conn.ssh_cmd
            ssh_sock_cmd = ssh_cmd[0:-1]
            ssh_user_host = remote.ssh_user_host
            check_cmd = ssh_sock_cmd + ["-O", "check", ssh_user_host]
            exit_cmd = ssh_sock_cmd + ["-O", "exit", ssh_user_host]

            # Confirm master is running.
            check_proc1 = subprocess.run(check_cmd, stdout=PIPE, stderr=PIPE, text=True)
            self.assertEqual(0, check_proc1.returncode, check_proc1.stderr)
            self.assertIn("Master running", check_proc1.stderr)

            # Kill the master to simulate an intermittent failure.
            subprocess.run(exit_cmd, stdout=PIPE, stderr=PIPE, text=True)
            check_proc2 = subprocess.run(check_cmd, stdout=PIPE, stderr=PIPE, text=True)
            self.assertNotEqual(0, check_proc2.returncode)
            self.assertIn("Control socket connect", check_proc2.stderr)

            last_refresh_before = conn._last_refresh_time
            with patch.object(conn, "_is_ssh_control_socket_usable", return_value=True):
                # Simulate a detection failure: refresh takes the fast path, ssh falls back to direct connection and succeeds.
                proc2 = conn.run_ssh_command(["echo", "two"], job=job, stdout=PIPE, stderr=PIPE, text=True, check=True)
                self.assertEqual("two\n", proc2.stdout)
                check_proc3 = subprocess.run(check_cmd, stdout=PIPE, stderr=PIPE, text=True)
                self.assertNotEqual(0, check_proc3.returncode)  # assert master is not running

                # Wait long enough so that a subsequent refresh performs -O check and recreates the master.
                time.sleep(remote.ssh_control_persist_secs - remote.ssh_control_persist_margin_secs + 1)

                proc3 = conn.run_ssh_command(["echo", "three"], job=job, stdout=PIPE, stderr=PIPE, text=True, check=True)
                self.assertEqual("three\n", proc3.stdout)
            check_proc4 = subprocess.run(check_cmd, stdout=PIPE, stderr=PIPE, text=True)
            self.assertEqual(0, check_proc4.returncode, check_proc4.stderr)
            self.assertIn("Master running", check_proc4.stderr)
            self.assertGreater(conn._last_refresh_time, last_refresh_before)

            # Replace the Unix domain socket file with a regular empty file, which will make _is_ssh_control_socket_usable()
            # return False and trigger the stale socket unlinking logic.
            socket_path = conn._ssh_socket_path
            self.assertIsNotNone(socket_path)
            socket_path = cast(str, socket_path)
            os.unlink(socket_path)
            with open(socket_path, "w", encoding="utf-8"):
                pass
            proc4 = conn.run_ssh_command(
                ["echo", "four-after-stale-socket-file"], job=job, stdout=PIPE, stderr=PIPE, text=True, check=True
            )
            self.assertEqual("four-after-stale-socket-file\n", proc4.stdout)
            check_proc5 = subprocess.run(check_cmd, stdout=PIPE, stderr=PIPE, text=True)
            self.assertEqual(0, check_proc5.returncode, check_proc5.stderr)
        finally:
            pool.shutdown()
